#!/usr/bin/python

# clipytool is a command-line utility for interacting with
#  a Cleanflight control board.  Configuration settings may be
#  backed up, restored and verified.  An interactive-terminal
#  mode is also implemented.
#
# Developed by etheli (ethomas997)
#  building on "cli.py" program by 'hrrr'

import os
import os.path
import sys
import stat
import serial
import io
import argparse
import time
import datetime
import glob

REVISION_STR = "clipytool version 1.1"

WINDOWS_DEF_PORT = "COM7"         # default serial port (if Windows)
NONWINDOWS_DEF_PORT = "/dev/tty.SLAB_USBtoUART"       # (if non-Windows)
DEF_BAUD_RATE = 115200

DEF_FNPRE_STR = "cli_bkp_"        # prefix for default filenames
DEF_FNSUF_STR = ".txt"            # suffix for default filenames

# list of commands that do not generate a text response
noRespCommandsList = ["led", "color", "aux", "adjrange", "servo"]

def readSerLine(init):
    global serLineTracker
    str = ""
    while True:
        ch = ser.read()
        if(ch == '\n'):
            if len(str) > 0:
                serLineTracker -= 1    # decrement on line received
            break
        if(ch == ''):
            break
        if(ch != '\r'):  
            str += ch
    return init + str

def getAllSerLines(f, dontPrintStr="", verifyFlag=False, init="# "):
    global numItemsProcessed
    while True:
        if (not f) and (ser.inWaiting() <= 0):   # if not saving to file then
            break                                # just return if data not ready
        line = readSerLine(init)
        init = ""
        if (line == "") and (ser.inWaiting() <= 0):
            break
        if (line != "# ") and (line != ""):
            if (not verifyFlag) and ((dontPrintStr == "") or (not line.startswith(dontPrintStr))):
                print line
            if f:
                if verifyFlag:
                    verifyLine(f,line)
                else:
                    f.write(line + '\n')
                    numItemsProcessed += 1

def sendSerial(str, f, verifyFlag=False):
    global serLineTracker
    global args
    global changesLastSerCmd
    if verifyFlag:
        print str.strip()
        if (args.changes is not None) and (not str.startswith("dump ")):
             changesLastSerCmd = str.strip()   # used if changes found in verifyLine()
    ser.write(str)
         # check if response to command is one line or two:
    if str.partition(' ')[0] in noRespCommandsList:
        serLineTracker += 1           # 1 for echo, no response
    else:
        serLineTracker += 2           # 1 for echo and 1 for response
    getAllSerLines(f,"",verifyFlag)

def getProfileValue(defStr):
    str = "profile"
    ser.write(str + '\n')
    readSerLine('')
    respStr = readSerLine('')   # response should be "profile #\n"
    time.sleep(0.1*sleepScale)
    getAllSerLines(0)
    str = str + ' '
    if respStr.startswith(str):
        return respStr[len(str):].strip()
    return defStr

def doBackupCommands(f, verifyFlag):
    global args
    global changesf
    savedProfileValue = getProfileValue("0")
    sendSerial('dump master\n', f, verifyFlag)
    sendSerial('profile 0\n', f, verifyFlag)
    time.sleep(0.1*sleepScale)
    sendSerial('dump profile\n', f, verifyFlag)
    sendSerial('profile 1\n', f, verifyFlag)
    time.sleep(0.1*sleepScale)
    sendSerial('dump profile\n', f, verifyFlag)
    sendSerial('profile 2\n', f, verifyFlag)
    time.sleep(0.1*sleepScale)
    sendSerial('dump profile\n', f, verifyFlag)
                             # restore profile to original:
    savedProfCmd = "profile " + savedProfileValue + '\n'
    if args.changes is not None:
        changesf.write(savedProfCmd)
    sendSerial(savedProfCmd,0) 
    print readSerLine('')
    readSerLine('')
    time.sleep(0.1*sleepScale)
    getAllSerLines(0,"",verifyFlag)
    sendSerial('rateprofile 0\n', f, verifyFlag)
    sendSerial('dump rates\n', f, verifyFlag)
    sendSerial('rateprofile 1\n', f, verifyFlag)
    sendSerial('dump rates\n', f, verifyFlag)
    sendSerial('rateprofile 2\n', f, verifyFlag)
    sendSerial('dump rates\n', f, verifyFlag)
    time.sleep(0.25*sleepScale)
    getAllSerLines(0)
    sendSerial("exit\n",0)   # don't leave it in CLI mode
    time.sleep(0.25*sleepScale)
    getAllSerLines(0, "Leaving CLI mode")
    return savedProfileValue

def verifyLine(f,recvStr):
    global numItemsMatched
    global numItemsMismatched
    global verifyListIndex
    global verifyLastMatchIdx
    global args
    global changesf
    if (args.changes is not None) and (recvStr.startswith("feature ")):
        changesf.write(recvStr + '\n')      # always write feature lines to changes file
    if verifyListIndex < len(f):
        while True:
            readStr = f[verifyListIndex].strip()
            verifyListIndex += 1
            if (len(readStr.strip()) > 0) or (verifyListIndex >= len(f)):
                break   # use line if not empty; otherwise skip
    else:
        readStr = ""
    if readStr == recvStr:
        numItemsMatched += 1
        verifyLastMatchIdx = verifyListIndex - 1
    else:
        checkIdx = verifyListIndex
        while True:     # check if mismatch because item(s) missing on controller
            if (checkIdx >= len(f)) or (checkIdx - verifyListIndex > 25):
                matchFlag = False
                break
            testStr = f[checkIdx].strip()
            if testStr.startswith("profile ") or testStr.startswith("rateprofile "):
                matchFlag = False      # don't scan past profile change
                break
            if testStr == recvStr:     # found match further down in file
                matchFlag = True
                break
            checkIdx += 1
        if matchFlag:
            numItemsMismatched += 1         # show initial mismatch
            showVerifyItem(recvStr,readStr)
            while verifyListIndex < checkIdx:
                numItemsMismatched += 1     # show any other mismatches
                showVerifyItem(recvStr,f[verifyListIndex].strip())
                verifyListIndex += 1
            verifyLastMatchIdx = verifyListIndex
            verifyListIndex += 1
            numItemsMatched += 1
        else:           # check if mismatch because item(s) missing in file
            checkIdx = verifyListIndex - 1
            while True:
                checkIdx -= 1     # don't check <= last match or too many back
                if (checkIdx <= verifyLastMatchIdx) or (verifyListIndex - checkIdx > 25):
                    matchFlag = False
                    break
                testStr = f[checkIdx].strip()
                if testStr == recvStr:      # found match earlier in file
                    matchFlag = True
                    break
            if matchFlag:
                verifyListIndex = checkIdx + 1   # resync index to match
                numItemsMatched += 1
                verifyLastMatchIdx = verifyListIndex - 1
        if not matchFlag:                   # no new matches were found
            numItemsMismatched += 1         # show single mismatch item
            showVerifyItem(recvStr,readStr)

def showVerifyItem(recvStr,readStr):
    global args
    global changesLastSerCmd
    global changesWriteCount
    global changesf
    if args.changes is None:
        print "  Mismatch - Received:  " + recvStr
        print "             Expected:  " + readStr
    else:
        print recvStr
        if (not recvStr.startswith("#")) and (not recvStr.startswith("feature ")):
            if len(changesLastSerCmd) > 0:       # feature lines handled above
                   # (ignore received string if duplicate of profile-select cmd)
                if recvStr != changesLastSerCmd:
                    changesf.write(changesLastSerCmd + '\n')  # enter profile-select cmd
                    changesf.write(recvStr + '\n')
                    changesWriteCount += 1
                    changesLastSerCmd = ""
            else:
                changesf.write(recvStr + '\n')
                changesWriteCount += 1

def openSerPort():
    global args
    global ser
    print (REVISION_STR + ":  Using serial port '" + str(args.port) +
          "' at " + str(args.baudrate) + " baud")
    try:
        toVal = 0.1
        if sleepScale > 1.0:          # if increasing sleep delays then
            toVal *= sleepScale       # also apply increase to port timeout
        ser = serial.Serial(args.port, args.baudrate, timeout=toVal)
        return True
    except:
        print("Error:  Cannot open port:  " + args.port)
        return False

def enterCliMode(delayFlag):
    cnt = 0
    while cnt < 5:
        if delayFlag:
            time.sleep(2.0)
        ser.write('#\n')
        readSerLine('')         # don't print "Entering CLI Mode..."  line(s)
        time.sleep(0.1*sleepScale)
        getAllSerLines(0, "# Entering CLI Mode")
        ser.write('version\n')         # verify in CLI
        str = readSerLine('')
        verStr = readSerLine('')
        str += verStr
        time.sleep(0.1*sleepScale)
        getAllSerLines(0, "#")
        if len(str) > 0:
            return verStr
        delayFlag = True
        cnt += 1
    print "Unable to communicate with controller"
    return False

def doVerifyItems(verifyFileName):
    print "Verifying items vs file '" + verifyFileName + "'"
    with open(verifyFileName,'r') as fHdl:
        f = fHdl.readlines()           # read file to list
    doBackupCommands(f,True)
    print "Number of items matched:  " + str(numItemsMatched)
    print "Number of items mismatched:  " + str(numItemsMismatched)
    if numItemsMismatched > 0:
        return False
    return True

def autoGenerateBackupFName(dirStr=""):
    nStr = DEF_FNPRE_STR + datetime.datetime.now().strftime("%Y%m%d_%H%M%S") + DEF_FNSUF_STR
    if (dirStr == ".") or (len(dirStr) <= 0):
        return nStr
    return os.path.join(dirStr,nStr)

def getNewestBackupFile(dirStr=""):
    if len(dirStr) > 0:
        listDir = dirStr
    else:
        listDir = "."
    # get all entries in the directory w/ stats
    entries = (os.path.join(listDir, fn) for fn in os.listdir(listDir))
    entries = ((os.stat(path), path) for path in entries)
    # leave only matching regular files, insert modification date
    entries = ((filstat[stat.ST_MTIME], filpath)
               for filstat, filpath in entries
               if (stat.S_ISREG(filstat[stat.ST_MODE]) and
               os.path.basename(filpath).startswith(DEF_FNPRE_STR) and
               os.path.basename(filpath).endswith(DEF_FNSUF_STR)))
    # make newest file be first item
    entries = sorted(entries, reverse=True)
    # return first item (with dir if not ".")
    if len(entries) > 0:
        retStr = os.path.basename(entries[0][1])
        if (dirStr != ".") and (len(dirStr) > 0):
            retStr = os.path.join(dirStr,retStr)
        return retStr
    return ""

def listSerialPorts():
    """Lists serial ports
    From:  http://stackoverflow.com/questions/12090503/listing-available-com-ports-with-python
    :raises EnvironmentError:
        On unsupported or unknown platforms
    """
    if sys.platform.startswith('win'):
        ports = ['COM' + str(i + 1) for i in range(256)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this is to exclude your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')
    plist = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            plist.append(port)
        except (OSError, serial.SerialException):
            pass
    if len(plist) > 0:
        print "Available serial ports:"
        for s in plist:
            print s
    else:
        print "No available serial ports found"

def doSerTerminal(verStr):
    print verStr             # show version string reported by controller
    while True:
        try:
            inStr = raw_input("# ")    # get line of input from user
        except KeyboardInterrupt:      # handle Ctrl-C
            print "quit"
            break
        if len(inStr) > 0:
            if inStr.lower() == "quit":
                break                  # exit without rebooting controller
            ser.flush()
            ser.write(inStr + '\n')    # send command to controller
            readSerLine('')            # don't display echo of command
            while True:                # display response to command
                time.sleep(0.1*sleepScale)
                if ser.inWaiting() <= 0:
                    break
                getAllSerLines(0, init="")
            if (inStr.lower() == "exit") or (inStr.lower() == "save"):
                break        # exit after sending cmd to reboot controller

def exitProgram(exitCode=0):
    global args
    try:
        if args.pauseonexit or (args.pauseonerror and exitCode != 0):
            raw_input("Press <Enter> to continue...")
    except:
        pass
    sys.exit(exitCode)

if os.name == "nt":
     portDefStr = WINDOWS_DEF_PORT          # default port, Windows
else:
     portDefStr = NONWINDOWS_DEF_PORT       # default port, non-Windows
parser = argparse.ArgumentParser(epilog=(REVISION_STR + "\r\n" +
                     "If backup filename not specified then defaults to \"" +
                     DEF_FNPRE_STR + "YYYYMMDD_hhmmss" + DEF_FNSUF_STR +
                     "\"\nIf restore/verify filename not specified then " +
                     "defaults to newest file matching \"" + DEF_FNPRE_STR +
                     "*" + DEF_FNSUF_STR +
                     "\"\nIf communications too fast for controller, " +
                     "try increasing 'sleepscale' > 1.0"),
                     formatter_class=argparse.RawTextHelpFormatter)
group = parser.add_mutually_exclusive_group()
group.add_argument("-b", "--backup", const=".", nargs='?', help="Save CLI settings to file: BACKUP")
group.add_argument("-r", "--restore", const=".", nargs='?', help="Restore CLI settings from file: RESTORE")
group.add_argument("-y", "--verify", const=".", nargs='?', help="Verify CLI settings vs file: VERIFY")
parser.add_argument("-c", "--changes", const=".", nargs='?', help="Save 'verify' changes to file: CHANGES")
parser.add_argument("-n", "--noverify", action='store_true', help="Don't perform verify after backup/restore")
parser.add_argument("-p", "--port", default=portDefStr, help=("Serial port to use (default=" + portDefStr + ")"))
parser.add_argument("-a", "--baudrate", default=DEF_BAUD_RATE, help=("Baud rate to use (default=" + str(DEF_BAUD_RATE) + ")"))
parser.add_argument("-s", "--sleepscale", default=1.0, help="Scale comm sleep delays (default=1.0)")
parser.add_argument("-u", "--pauseonerror", action='store_true', help="Pause before exit if error")
parser.add_argument("-x", "--pauseonexit", action='store_true', help="Pause before exit (always)")
parser.add_argument("-l", "--listports", action='store_true', help="Display available serial ports")
parser.add_argument("-t", "--terminal", action='store_true', help="Enter interactive terminal mode")
parser.add_argument("-v", "--version", action='version', version=REVISION_STR)
args = parser.parse_args()
serLineTracker = 0           # increment on send, decrement on receive
numItemsProcessed = 0
numItemsMatched = 0
numItemsMismatched = 0
verifyListIndex = 0
verifyLastMatchIdx = -1
processFileName = ""
changesFileName = ""
changesLastSerCmd = ""
changesWriteCount = 0
sleepScale = float(args.sleepscale)

#list serial ports
if args.listports:
    print REVISION_STR
    listSerialPorts()
    exitProgram()

#interactive terminal
if args.terminal:
    if not openSerPort():
        exitProgram(1)
    retVal = enterCliMode(False)
    if not retVal:
        exitProgram(1)
    doSerTerminal(retVal)
    exitProgram()

#must specify -y when -c specified
if (args.changes is not None) and (args.verify is None):
    print "Error:  Cannot specify -c without -y"
    parser.print_help()
    exitProgram(1)   

#check at least one option is used
if (args.backup is None) and (args.restore is None) and (args.verify is None):
    print "Error:  Either -b, -r  or -y must be specified"
    parser.print_help()
    exitProgram(1)   

#check backup file does not exist
if args.backup is not None:
    processFileName = args.backup
    if (len(processFileName) <= 0) or os.path.isdir(processFileName):
        processFileName = autoGenerateBackupFName(processFileName)
    if os.path.isfile(processFileName):
        print "Error:  File '" + processFileName + "' exists"
        print "Delete it or backup to a different file"
        exitProgram(1)

#check restore file does exist
if args.restore is not None:
    processFileName = args.restore
    if (len(processFileName) <= 0) or os.path.isdir(processFileName):
        processFileName = getNewestBackupFile(processFileName)
        if len(processFileName) <= 0:
            print "Error:  Restore file not specified and no matching file found"
            exitProgram(1)
    if (not(os.path.isfile(processFileName))) or (not(os.access(processFileName, os.R_OK))):
        print "Error:  Restore file '" + processFileName + "' cannot be read" 
        exitProgram(1)
    if os.stat(processFileName).st_size > 1000000L:
        print "Error:  Restore file '" + processFileName + "' too large" 
        exitProgram(1)

#check verify file does exist
if args.verify is not None:
    processFileName = args.verify
    if (len(processFileName) <= 0) or os.path.isdir(processFileName):
        processFileName = getNewestBackupFile(processFileName)
        if len(processFileName) <= 0:
            print "Error:  Verify file not specified and no matching file found"
            exitProgram(1)
    if (not(os.path.isfile(processFileName))) or (not(os.access(processFileName, os.R_OK))):
        print "Error:  Verify file '" + processFileName + "' cannot be read" 
        exitProgram(1)
    if os.stat(processFileName).st_size > 1000000L:
        print "Error:  Verify file '" + processFileName + "' too large" 
        exitProgram(1)

#check changes file does not exist
if args.changes is not None:
    changesFileName = args.changes
    if (len(changesFileName) <= 0) or os.path.isdir(changesFileName):
        changesFileName = autoGenerateBackupFName(changesFileName)
    if os.path.isfile(changesFileName):
        print "Error:  File '" + changesFileName + "' exists"
        print "Delete it or write changes to a different file"
        exitProgram(1)

#open serial port
if not openSerPort():
    exitProgram(1)

#enter CLI mode
if not enterCliMode(False):
    exitProgram(1)

#backup settings
if args.backup is not None:
    print "Backing up items to file '" + processFileName + "'"
    f = open(processFileName, 'w')
    profVal = doBackupCommands(f,False)
    f.write("profile " + profVal + '\n')    # append command to restore profile
    f.close() 
    print "Number of items backed up:  " + str(numItemsProcessed)
    print "Backed up items to file '" + processFileName + "'"
    if numItemsProcessed <= 0:
        exitProgram(1)
    if not args.noverify:
        time.sleep(2.0)
        print " ----"
        if not enterCliMode(True):
            exitProgram(1)
        if not doVerifyItems(processFileName):
            exitProgram(2)

#restore settings
if args.restore is not None:
    print "Restoring items from file '" + processFileName + "'"
    f = open(processFileName, 'r')
    prevLine = ""
    cnt = 0
    for line in f:
        if len(line.strip()) > 0:
            numItemsProcessed += 1
            if ((line[0] !='#') and (not ("Cleanflight" in line)) and (line != prevLine)): 
                sendSerial(line,0)
                prevLine = line
                cnt += 1
                if line.startswith("profile "):
                    time.sleep(0.25*sleepScale)  # allow time to process 'profile' cmd
                    cnt = 0
                    if serLineTracker < 0:       # if receive ahead of send then
                        serLineTracker = 0       # reset tracker
                elif cnt > 25:
                    time.sleep(0.1*sleepScale)   # pause at least every 25 lines
                    cnt = 0
                    if serLineTracker < 0:       # if receive ahead of send then
                        serLineTracker = 0       # reset tracker
                elif serLineTracker > 9:    # if send getting ahead of receive
                    time.sleep(0.05*sleepScale)  # pause to let data arrive
                    readSerLine('')              # read received data
                    getAllSerLines(0)
                    if serLineTracker > 9:       # if send still ahead of receive then
                        serLineTracker -= 10     # reduce tracker to prevent looping
                    elif serLineTracker < 0:     # if receive ahead of send then
                        serLineTracker = 0       # reset tracker
    f.close()
    if (numItemsProcessed > 0) and (prevLine.startswith("profile ")):
      numItemsProcessed -= 1      # don't include extra line for restoring profile #
    time.sleep(0.25*sleepScale)
    getAllSerLines(0)
    sendSerial("save\n",0)
    time.sleep(0.25*sleepScale)
    getAllSerLines(0)
    print "Number of items restored:  " + str(numItemsProcessed)
    print "Restored items from file '" + processFileName + "'"
    if numItemsProcessed <= 0:
        exitProgram(1)
    if not args.noverify:
        time.sleep(2.0)
        print " ----"
        if not enterCliMode(True):
            exitProgram(1)
        if not doVerifyItems(processFileName):
            exitProgram(2)

#verify settings
if args.verify is not None:
    if args.changes is not None:
        print "Writing changes to file '" + changesFileName + "'"
        changesf = open(changesFileName, 'w')
    retVal = doVerifyItems(processFileName)
    if args.changes is not None:
        changesf.close()
        if changesWriteCount > 0:
            print "Number of changes written to file '" + changesFileName + "':  " + str(changesWriteCount)
        else:
            print "No changes, file '" + changesFileName + "' not generated"
            os.remove(changesFileName)
    if not retVal:
        exitProgram(2)

ser.close()
exitProgram()
